## Helper functions for evaluating Fourier, Chebyshev things
function FourierMat(θ::AbstractVector, Nθ::Integer)
  A = zeros(length(θ), 2Nθ+1);
  A[:, 1:2:end] = cos.(θ * (0:Nθ)');
  A[:, 2:2:end] = sin.(θ * (1:Nθ)');

  return A
end

function dFourierMat(θ::AbstractVector, Nθ::Integer)
  A = zeros(length(θ), 2Nθ+1);
  A[:, 1:2:end] = - sin.(θ * (0:Nθ)')*Diagonal(0:Nθ);
  A[:, 2:2:end] = cos.(θ * (1:Nθ)')*Diagonal(1:Nθ);

  return A
end

function ddFourierMat(θ::AbstractVector, Nθ::Integer)
  A = zeros(length(θ), 2Nθ+1);
  A[:, 1:2:end] = - cos.(θ * (0:Nθ)')*Diagonal((0:Nθ).^2)
  A[:, 2:2:end] = - sin.(θ * (1:Nθ)')*Diagonal((1:Nθ).^2)

  return A
end


function FourierGridMat(Nθ::Integer; backward::Bool=false)
  θ = (0:2Nθ+1) .* (2π/(2Nθ+2))
  A = FourierMat(θ, Nθ);

  if backward
    for n = 1:2Nθ+1
      A[:, n] = A[:,n] ./ (A[:,n]'*A[:,n])
    end
  end

  return A;
end

function ChebyshevMat(θ::AbstractVector, Nζ::Integer)
  return cos.(θ * (0:Nζ)');
end

function dChebyshevMat(θ::AbstractVector, Nζ::Integer)
  A = zeros(length(θ), Nζ+1);
  i_0 = (θ .< 1e-8)
  i_π = (θ .> π-1e-8)
  i_e = (.!i_0) .&& (.!i_π)

  n = 0:Nζ
  A[i_0, :] .= ones(sum(i_0)) * (n.^2)'
  A[i_π, :] .= ones(sum(i_π)) * (-(-1).^n .* n.^2)'
  A[i_e, :] .= Diagonal(csc.(θ[i_e])) * sin.(θ[i_e] * n') * Diagonal(n);

  return A
end

function ddChebyshevMat(θ::AbstractVector, Nζ::Integer)
  A = zeros(length(θ), Nζ+1);

  i_0 = (θ .< 1e-8)
  i_π = (θ .> π-1e-8)
  i_e = (.!i_0) .&& (.!i_π)

  n = 0:Nζ

  v = -(1/3) .* n.^2 .* (1 .- n.^2);
  A[i_0, :] .= ones(sum(i_0)) * v'

  v =  (-1).^n .* v
  A[i_π, :] .= ones(sum(i_π)) * v'

  cscθ = csc.(θ[i_e]);
  A[i_e, :] .= -Diagonal(cscθ.^2) * cos.(θ[i_e] * n') * Diagonal(n.^2)
  A[i_e, :] .= A[i_e, :] + Diagonal(cscθ.^3 .* cos.(θ[i_e])) * sin.(θ[i_e] * n') * Diagonal(n);

  return A;
end


function ChebyshevGridMat(Nζ::Integer; backward::Bool=false)
  A = ChebyshevMat(LinRange(0, π, Nζ+1), Nζ)

  if backward
    # Quadrature rule
    w = ones(Nζ+1)
    w[1] = 1/2; w[end] = 1/2;
    W = Diagonal(w);

    for n = 1:Nζ+1
      A[:, n] = W*A[:,n] ./ (A[:,n]'*W*A[:,n])
    end
  end

  return A;
end

## Get an interpolant of a surface.
#   Nζ is the degree of the polynomial interpolant minus 1
#   Nα is the degree of the Fourier interpolant
#   specify c2_nodes, κ_nodes, bsuppar_nodes only if you already have the
#     coefficients on a grid (used in adaptive method)
function BallooningCoefInterp(vmecSurface::VmecSurface, Nζ::Integer, Nα::Integer;
                              c2_nodes=[], κ_nodes = [], bsuppar_nodes = [])
  nfp = vmecSurface.nfp;
  iota = vmecSurface.iota[1];

  if c2_nodes == []
    ζs = (π/nfp) .* (-cos.(LinRange(0, π, Nζ+1)) .+ 1);
    αvec = (0:2Nα+1) .* (2π/(2Nα+2)); # only really need 2Nα+1 points, but an additional is used for nesting
    c2_nodes, κ_nodes, bsuppar_nodes = ballooning_coefficient_grid(vmecSurface, ζs, αvec)
  end

  AFinvT = FourierGridMat(Nα; backward=true)
  ACinvT = ChebyshevGridMat(Nζ; backward=true)

  c2 = zeros(1 + 2Nα, Nζ + 1, 3);
  for ii = 1:3
    c2[:,:,ii] = AFinvT' * c2_nodes[:, :, ii] * ACinvT;
  end

  κ = zeros(1 + 2Nα, Nζ + 1, 2);
  for ii = 1:2
    κ[:,:,ii] = AFinvT' * κ_nodes[:, :, ii] * ACinvT;
  end

  bsuppar = AFinvT' * bsuppar_nodes * ACinvT;

  BallooningCoefInterp(c2, κ, bsuppar, nfp, iota);
end

function scale_bci(bci::BallooningCoefInterp, vmecSurface::VmecSurface, params::BalloonParams)
  phipc = vmecSurface.phi[2]/(-2*π) #phi derivative
  μ0 = 4*π*1.0e-7
  prespn = 2*vmecSurface.pres[2]*μ0/(params.beta0*params.b0_v^2);     # pressure normalized to p_0
  phipn2 = (phipc/(params.b0_v*params.amin^2))^2;      # perpend. lengths normalized to a
  eps2 = (params.r0/params.amin)^2;
  factor = eps2*params.beta0*prespn/phipn2;

  bci.c2      .= bci.c2 .* ((params.amin^2) * (params.b0_v^2));  # perpend. wave vector squared over mag field
  bci.bsuppar .= params.r0 .* bci.bsuppar ./ params.b0_v;
  bci.κ       .= factor .* bci.κ
end

function BallooningCoefInterp(vmecSurface::VmecSurface, Nζ::Integer, Nα::Integer, params::BalloonParams)
  bci = BallooningCoefInterp(vmecSurface, Nζ, Nα)
  scale_bci(bci, vmecSurface, params)
  return bci
end

function get_err_on_grid(bci, c2_nodes, κ_nodes, bsuppar_nodes, α, ζ)
  c2_α_interp, κ_α_interp, bsuppar_α_interp = eval_coefs(bci, α[2:2:end], ζ[1:2:end])

  c2_α_err = c2_nodes[2:2:end, 1:2:end,:] - c2_α_interp
  κ_α_err = κ_nodes[2:2:end, 1:2:end,:] - κ_α_interp
  bsuppar_α_err = bsuppar_nodes[2:2:end, 1:2:end] - bsuppar_α_interp

  c2_α_relerr = norm(c2_α_err)/norm(c2_nodes[2:2:end, 1:2:end,:])
  κ_α_relerr = norm(κ_α_err)/norm(κ_nodes[2:2:end, 1:2:end,:])
  bsuppar_α_relerr = norm(bsuppar_α_err)/norm(bsuppar_nodes[2:2:end, 1:2:end])

  c2_ζ_interp, κ_ζ_interp, bsuppar_ζ_interp = eval_coefs(bci, α[1:2:end], ζ[2:2:end])

  c2_ζ_err = c2_nodes[1:2:end, 2:2:end,:] - c2_ζ_interp
  κ_ζ_err = κ_nodes[1:2:end, 2:2:end,:] - κ_ζ_interp
  bsuppar_ζ_err = bsuppar_nodes[1:2:end, 2:2:end] - bsuppar_ζ_interp

  c2_ζ_relerr = norm(c2_ζ_err)/norm(c2_nodes[1:2:end, 2:2:end, :])
  κ_ζ_relerr = norm(κ_ζ_err)/norm(κ_nodes[1:2:end, 2:2:end, :])
  bsuppar_ζ_relerr = norm(bsuppar_ζ_err)/norm(bsuppar_nodes[1:2:end, 2:2:end])

  return [c2_α_relerr, κ_α_relerr, bsuppar_α_relerr], [c2_ζ_relerr, κ_ζ_relerr, bsuppar_ζ_relerr]
end

## Adaptive interpolation
#   Reuses previously computed grid points, so the number of evaluations of
#   field lines is kept to a minimum. Refines in α and ζ differently, so
#   it is possible to have a lower-order interpolant in α or ζ
function BallooningCoefInterp(vmecSurface::VmecSurface; reltol::Number=1e-4)
  errF = 1;
  errC = 1;
  Nα = 10;
  Nζ = 10;
  Nrefinement=4;
  nfp = vmecSurface.nfp;
  iota = vmecSurface.iota[1];

  # Get interpolant on initial grid
  ζs = (π/nfp) .* (-cos.(LinRange(0, π, Nζ+1)) .+ 1);
  αvec = (0:2Nα+1) .* (2π/(2Nα+2));
  c2_nodes, κ_nodes, bsuppar_nodes = ballooning_coefficient_grid(vmecSurface, ζs, αvec)
  bci = BallooningCoefInterp(vmecSurface, Nζ, Nα; c2_nodes, κ_nodes, bsuppar_nodes)

  αdoubled = true; ζdoubled = true;
  for ii = 1:Nrefinement
    # Refine the grid
    c2_nodes_next = zeros(4Nα+4, 2Nζ+1, 3);
    κ_nodes_next = zeros(4Nα+4, 2Nζ+1, 2);
    bsuppar_nodes_next = zeros(4Nα+4, 2Nζ+1);
    α = (0:4Nα+3) .* (2π/(4Nα+4));
    θ = LinRange(0, π, 2Nζ+1);
    ζ = (π/nfp) .* (-cos.(θ) .+ 1);

    # Fill refined grid with already known coefficients
    if αdoubled && ζdoubled
      c2_nodes_next[1:2:end, 1:2:end, :] = c2_nodes;
      κ_nodes_next[1:2:end, 1:2:end, :] = κ_nodes;
      bsuppar_nodes_next[1:2:end, 1:2:end] = bsuppar_nodes;
    elseif αdoubled
      c2_nodes_next[1:2:end, :, :] = c2_nodes;
      κ_nodes_next[1:2:end, :, :] = κ_nodes;
      bsuppar_nodes_next[1:2:end, :, :] = bsuppar_nodes;
    else
      c2_nodes_next[:, 1:2:end, :] = c2_nodes;
      κ_nodes_next[:, 1:2:end, :] = κ_nodes;
      bsuppar_nodes_next[:, 1:2:end, :] = bsuppar_nodes;
    end
    c2_nodes, κ_nodes, bsuppar_nodes = c2_nodes_next, κ_nodes_next, bsuppar_nodes_next;

    # Compute needed coefficients to calculate error
    if αdoubled
      c2_α, κ_α, bsuppar_α = ballooning_coefficient_grid(vmecSurface, ζ[1:2:end], α[2:2:end])
      c2_nodes[2:2:end, 1:2:end, :] = c2_α;
      κ_nodes[2:2:end, 1:2:end, :] = κ_α;
      bsuppar_nodes[2:2:end, 1:2:end] = bsuppar_α;
    end

    if ζdoubled
      c2_ζ, κ_ζ, bsuppar_ζ = ballooning_coefficient_grid(vmecSurface, ζ[2:2:end], α[1:2:end])
      c2_nodes[1:2:end, 2:2:end, :] = c2_ζ;
      κ_nodes[1:2:end, 2:2:end, :] = κ_ζ;
      bsuppar_nodes[1:2:end, 2:2:end] = bsuppar_ζ;
    end

    # Evaluate error in α and ζ directions
    αerrs, ζerrs = get_err_on_grid(bci, c2_nodes, κ_nodes, bsuppar_nodes, α, ζ)

    # Check error to see if refinement is necessary in both directions
    αdoubled = maximum(αerrs) > reltol
    if αdoubled
      Nα = 2Nα + 1;
    end

    ζdoubled = maximum(ζerrs) > reltol
    if ζdoubled
      Nζ = 2Nζ
    end

    if !αdoubled && !ζdoubled
      return bci
    end

    # Get the rest of the values on the grid
    c2_αζ, κ_αζ, bsuppar_αζ = ballooning_coefficient_grid(vmecSurface, ζ[2:2:end], α[2:2:end]);
    c2_nodes[2:2:end, 2:2:end, :] = c2_αζ;
    κ_nodes[2:2:end, 2:2:end, :] = κ_αζ;
    bsuppar_nodes[2:2:end, 2:2:end] = bsuppar_αζ;

    #
    if αdoubled && ζdoubled
      bci = BallooningCoefInterp(vmecSurface, Nζ, Nα; c2_nodes, κ_nodes, bsuppar_nodes)
    elseif αdoubled
      bci = BallooningCoefInterp(vmecSurface, Nζ, Nα; c2_nodes=c2_nodes[:, 1:2:end, :],
                               κ_nodes=κ_nodes[:, 1:2:end, :], bsuppar_nodes=bsuppar_nodes[:, 1:2:end])
    else
      bci = BallooningCoefInterp(vmecSurface, Nζ, Nα; c2_nodes=c2_nodes[1:2:end, :, :],
                               κ_nodes=κ_nodes[1:2:end, :, :], bsuppar_nodes=bsuppar_nodes[1:2:end, :])
    end
  end

  return bci
end

function BallooningCoefInterp(vmecSurface::VmecSurface, params::BalloonParams; reltol::Number=1e-4)
  bci = BallooningCoefInterp(vmecSurface; reltol)
  scale_bci(bci, vmecSurface, params)
  return bci
end

function get_c2(bci::BallooningCoefInterp)
  return bci.c2
end

function get_κ(bci::BallooningCoefInterp)
  return bci.κ
end

function get_bsuppar(bci::BallooningCoefInterp)
  return bci.bsuppar
end

function get_fα(bci::BallooningCoefInterp)
  return bci.fα;
end

function get_Nα(bci::BallooningCoefInterp)
  return size(get_bsuppar(bci), 1) ÷ 2
end

function get_Nζ(bci::BallooningCoefInterp)
  return size(get_bsuppar(bci), 2)-1
end

function get_nfp(bci::BallooningCoefInterp)
  return bci.nfp;
end

function get_iota(bci::BallooningCoefInterp)
  return bci.iota;
end

## eval_coefs:
# Get the values of c2_coefs, κ_coefs, and bsuppar_coefs on a grid defined
# by α and ζ
function eval_coefs(bci::BallooningCoefInterp, α::AbstractVector, ζ::AbstractVector;
                    eval_c2::Bool=true, eval_κ::Bool=true, eval_bsuppar::Bool=true)
  # TODO: should add a check that ζ is in range
  AF = FourierMat(α, get_Nα(bci));
  θ = acos.(1 .- ζ./(π/get_nfp(bci)));
  AC = ChebyshevMat(θ, get_Nζ(bci));

  nα = length(α);
  nζ = length(ζ);

  c2_nodes = []; κ_nodes = []; bsuppar_nodes = [];

  if eval_c2
    c2_nodes = zeros(nα, nζ, 3)
    c2 = get_c2(bci);
    for ii = 1:3
      c2_nodes[:,:,ii] = (AF * c2[:,:,ii]) * AC';
    end
  end

  if eval_κ
    κ_nodes  = zeros(nα, nζ, 2)
    κ = get_κ(bci);
    for ii = 1:2
      κ_nodes[:,:,ii] = (AF * κ[:,:,ii]) * AC';
    end
  end

  if eval_bsuppar
    bsuppar = get_bsuppar(bci);
    bsuppar_nodes = (AF * bsuppar) * AC';
  end

  return c2_nodes, κ_nodes, bsuppar_nodes
end

function eval_coefs(bci::BallooningCoefInterp, α::Number, ζ::AbstractVector;
                      eval_c2::Bool=true, eval_κ::Bool=true, eval_bsuppar::Bool=true)
  c2_nodes, κ_nodes, bsuppar_nodes = eval_coefs(bci, [α], ζ; eval_c2, eval_κ, eval_bsuppar)
  nζ = length(ζ);
  return reshape(c2_nodes, nζ, 3), reshape(κ_nodes, nζ, 2), vec(bsuppar_nodes)
end


function ddeval_kernel(coefs, AF, dAF, ddAF, AC, dAC, ddAC)
  # TODO: should add a check that ζ is in range
  n1 = size(AF, 1); n2 = size(AC, 1);
  df = zeros(n1, n2, 2);
  ddf = zeros(n1, n2, 2, 2);

  f            = (AF * coefs) * AC';
  df[:,:,1]    = (dAF * coefs) * AC';
  df[:,:,2]    = (AF * coefs) * dAC';
  ddf[:,:,1,1] = (ddAF * coefs) * AC';
  ddf[:,:,2,1] = (dAF * coefs) * dAC';
  ddf[:,:,2,2] = (AF * coefs) * ddAC';
  ddf[:,:,1,2] = ddf[:,:,2,1];

  return f, df, ddf
end


## ddeval_coefs:
# Get the values of c2_coefs, κ_coefs, and bsuppar_coefs and their derivatives
# w.r.t. α and ζ on a grid defined by α and ζ
function ddeval_coefs(bci::BallooningCoefInterp, α::AbstractVector, ζ::AbstractVector;
                      eval_c2::Bool=true, eval_κ::Bool=true, eval_bsuppar::Bool=true)
  nfp = get_nfp(bci);
  L = 2π/nfp;
  AF = FourierMat(α, get_Nα(bci));
  dAF = dFourierMat(α, get_Nα(bci));
  ddAF = ddFourierMat(α, get_Nα(bci));
  θ = acos.(1 .- ζ./(π/get_nfp(bci)));
  AC   =               ChebyshevMat(θ, get_Nζ(bci));
  dAC  = (-2/L)   .*  dChebyshevMat(θ, get_Nζ(bci));
  ddAC = (-2/L)^2 .* ddChebyshevMat(θ, get_Nζ(bci));

  nα = length(α);
  nζ = length(ζ);

  c2, dc2, ddc2 = [], [], []
  if eval_c2
    c2 = zeros(nα, nζ, 3)
    dc2 = zeros(nα, nζ, 2, 3)
    ddc2 = zeros(nα, nζ, 2, 2, 3)
    c2_coef = get_c2(bci);
    for ii = 1:3
      c2[:,:,ii], dc2[:,:,:,ii], ddc2[:,:,:,:,ii] = (
                    ddeval_kernel(c2_coef[:,:,ii], AF, dAF, ddAF, AC, dAC, ddAC));
    end
  end

  κ, dκ, ddκ = [], [], []
  if eval_κ
    κ  = zeros(nα, nζ, 2)
    dκ = zeros(nα, nζ, 2, 2)
    ddκ = zeros(nα, nζ, 2, 2, 2)
    κ_coef = get_κ(bci);
    for ii = 1:2
      κ[:,:,ii], dκ[:,:,:,ii], ddκ[:,:,:,:,ii] = (
                     ddeval_kernel(κ_coef[:,:,ii], AF, dAF, ddAF, AC, dAC, ddAC));
    end
  end

  bsuppar, dbsuppar, ddbsuppar = [], [], []
  if eval_bsuppar
    bsuppar_coef = get_bsuppar(bci);
    bsuppar, dbsuppar, ddbsuppar = (
                       ddeval_kernel(bsuppar_coef, AF, dAF, ddAF, AC, dAC, ddAC));
  end

  return c2, κ, bsuppar, dc2, dκ, dbsuppar, ddc2, ddκ, ddbsuppar
end

function ddeval_coefs(bci::BallooningCoefInterp, α::Number, ζ::AbstractVector;
                      eval_c2::Bool=true, eval_κ::Bool=true, eval_bsuppar::Bool=true)
  c2, κ, bsuppar, dc2, dκ, dbsuppar, ddc2, ddκ, ddbsuppar = (
                       ddeval_coefs(bci, [α], ζ; eval_c2, eval_κ, eval_bsuppar))
  nζ = length(ζ);

  if eval_c2
    c2 = reshape(c2, nζ, 3)
    dc2 = reshape(dc2, nζ, 2, 3)
    ddc2 = reshape(ddc2, nζ, 2, 2, 3)
  end

  if eval_κ
    κ = reshape(κ, nζ, 2)
    dκ = reshape(dκ, nζ, 2, 2)
    ddκ = reshape(ddκ, nζ, 2, 2, 2)
  end

  if eval_bsuppar
    bsuppar = reshape(bsuppar, nζ)
    dbsuppar = reshape(dbsuppar, nζ, 2)
    ddbsuppar = reshape(ddbsuppar, nζ, 2, 2)
  end

  return c2, κ, bsuppar, dc2, dκ, dbsuppar, ddc2, ddκ, ddbsuppar
end

# Map alpha forward a sheet
#  (α, ζ) -> (map_α(bci, α), ζ-2π/nfp)
function map_α(bci::BallooningCoefInterp, α::AbstractVector)
  return α .+ (2π * get_iota(bci) / get_nfp(bci));
end

function map_α(bci::BallooningCoefInterp, α::Number)
  return map_α(bci, [α])[1]
end

# Map alpha backwards a sheet
#  (α, ζ) -> (map_α_inv(bci, α), ζ+2π/nfp)
function map_α_inv(bci::BallooningCoefInterp, α::AbstractVector)
  return α .- (2π * get_iota(bci) / get_nfp(bci));
end

function map_α_inv(bci::BallooningCoefInterp, α::Number)
  return map_α_inv(bci, [α])[1]
end
