include("../src/ratapprox.jl")


function test_ratapprox()
  N = 1
  r = RatFourier(N)
  set_num!(r, [0.0, 0.0, 1.0]);
  set_den!(r, [0.5, 0.0])

  θ = 0:π/50:2π
  z = eval(r, θ);

  return θ,z
end

function test_rat_derivs()
  N = 4
  r = RatFourier(N)
  set_num!(r, randn(2N+1));
  set_den!(r, rand(2N) .* 1e-1)


  # θ = 0:π/50:2π
  θ = randn()
  A = FourierMat([θ], N);
  dA = dFourierMat([θ], N);
  ddA = ddFourierMat([θ], N);
  z = eval(r, θ; A);
  dz = deval(r, θ; A, dA)
  ddz = ddeval(r, θ; A, dA, ddA);

  ϵ = 1e-4
  Δθ = ϵ .* randn()
  z2 = eval(r, θ + Δθ)
  dz2 = deval(r, θ + Δθ)

  println("z2 - z = $(z2-z), dz*Δθ = $(dz*Δθ), err = $(z2-z - dz*Δθ)")
  println("dz2 - dz = $(dz2-dz), ddz*Δθ = $(ddz*Δθ), err = $(dz2-dz-ddz*Δθ)")
end

function rat_interp_test()
  N = 4
  r = RatFourier(N)
  set_num!(r, randn(2N+1));
  set_den!(r, rand(2N) .* 1e-1)

  θ = (0:4N) .* (2π/(4N+1))
  z = eval(r, θ)
  r2 = interp_RatFourier(θ, z)

  println("Interp:")
  println(" norm(num-num2) = $(norm(get_num(r)-get_num(r2)))")
  println(" norm(den-den2) = $(norm(get_den(r)-get_den(r2)))")

  θ = (0:8N) .* (2π/(8N+1))
  z = eval(r, θ)
  r2 = lls_RatFourier(θ, z, N)
  println("Oversampled least squares:")
  println(" norm(num-num2) = $(norm(get_num(r)-get_num(r2)))")
  println(" norm(den-den2) = $(norm(get_den(r)-get_den(r2)))")
end

function rat_example(;k=0.0)
  f = (θ) -> sin(θ + k.*sin(θ))

  N = 5;
  Nθ = ((4N+1)*3)÷2
  θ = (0:Nθ-1) .* (2π/Nθ)
  r = lls_RatFourier(θ, f.(θ), N)

  println("nonlinear interp error for N=$N: $(norm(f.(θ) - eval(r, θ)))")
  return θ, f.(θ), eval(r,θ)
end


function rat_example2(;k=0.0, N=5)
  h = (θ) -> θ + k.*sin.(θ)
  hp = (θ) -> 1 + k.*cos.(θ)

  hinv = (θ) -> begin
    α = copy(θ)
    for ii = 1:10
      α = α - (h(α)-θ) ./ hp(α)
    end
    α
  end

  f = (θ) -> sin.(hinv(θ)) # .+ 1. .+ cos.(2 .*hinv(θ))

  Nθ = ((4N+1)*3)÷2
  θ = (0:Nθ-1) .* (2π/Nθ)
  r = lls_RatFourier(θ, f.(θ), N)
  println("nonlinear error for lls algorithm, N=$N: $(2π*norm(f.(θ) - eval(r, θ))/sqrt(Nθ))")

  r2 = nls_RatFourier(θ, f.(θ), N)
  println("nonlinear error for nls algorithm, N=$N: $(2π*norm(f.(θ) - eval(r2, θ))/sqrt(Nθ))")

  Fcoef = lls_Fourier(θ, f.(θ), N)
  AF = FourierMat(θ, 2N);
  println("Fourier series error w/ same number of unkowns: $(2π*norm(f.(θ) - AF*Fcoef)/sqrt(Nθ))")

  return θ, f.(θ), eval(r,θ), eval(r2,θ), AF*Fcoef
end

function nls_FDtest()
  N = 4
  r = RatFourier(N)
  set_num!(r, randn(2N+1));
  set_den!(r, rand(2N) .* 1e-1)


  Nθ = ((4N+1)*3)÷2
  θ = (0:Nθ-1) .* (2π/Nθ)
  z = randn(length(θ))

  resid = zeros(Nθ);
  resid2 = zeros(Nθ);
  J = zeros(Nθ, 4N+1);
  J2 = zeros(Nθ, 4N+1);
  H = zeros(4N+1, 4N+1)
  AF = FourierMat(θ, N);

  resid_RatFourier!(resid, r, θ, z, AF)
  jacob_RatFourier(J, r, θ, z, AF)
  hess_RatFourier(H, resid, J, r, θ, z2, AF)

  ϵ = 1e-4
  Δn = ϵ .* randn(2N+1)
  Δd = ϵ .* randn(2N)
  Δcoef = zeros(4N+1);
  Δcoef[1:2N+1] = Δn;
  Δcoef[2N+2:end] = Δd
  r2 = deepcopy(r);
  set_num!(r2, get_num(r2) + Δn)
  set_den!(r2, get_den(r2)[2:end] + Δd)

  resid_RatFourier!(resid2, r2, θ, z, AF)
  jacob_RatFourier(J2, r2, θ, z, AF)

  println("norm(resid2-resid) = $(norm(resid2-resid)), norm(resid2-resid-J*Δcoef) = $(norm(resid2-resid-J*Δcoef))")
  println("norm(J2'*resid2-J'*resid) = $(norm(J2'*resid2-J'*resid)), norm(J2'*resid2-J'*resid-H*Δcoef) = $(norm(J2'*resid2-J'*resid-H*Δcoef))")
end

function FourierInterpConvergence(;k=0.9,Nmax=1500)
  h = (θ) -> θ + k.*sin.(θ)
  hp = (θ) -> 1 + k.*cos.(θ)

  hinv = (θ) -> begin
    α = copy(θ)
    for ii = 1:10
      α = α - (h(α)-θ) ./ hp(α)
    end
    α
  end

  f = (θ) -> sin.(hinv(θ)) # .+ 1. .+ cos.(2 .*hinv(θ))


  Nθtest = 16Nmax+1
  θtest = (0:Nθtest-1) .* (2π/Nθtest)
  ztest = f.(θtest)

  NN = 50;

  logNs = LinRange(0, log10(Nmax), NN)
  Ns = [floor(Int64, 10 ^ (logN)) for logN in logNs]
  errs = zeros(NN);

  for ii = 1:NN
    N = Ns[ii]÷2;
    Nθ = 4N+1
    θ = (0:Nθ-1) .* (2π/Nθ)
    coef = lls_Fourier(θ, f.(θ), N)
    println("length(coef) = $(length(coef))")

    AF = FourierMat(θtest, 2N)
    errs[ii] = norm(AF*coef-ztest)/sqrt(Nθtest)
  end

  return Ns, errs
end


function LeastSquaresConvergence(;k=0.9, Nmin = 1, Nmax=750)
  h = (θ) -> θ + k.*sin.(θ)
  hp = (θ) -> 1 + k.*cos.(θ)

  hinv = (θ) -> begin
    α = copy(θ)
    for ii = 1:10
      α = α - (h(α)-θ) ./ hp(α)
    end
    α
  end

  f = (θ) -> sin.(hinv(θ)) # .+ 1. .+ cos.(2 .*hinv(θ))


  Nθtest = 16Nmax+1
  θtest = (0:Nθtest-1) .* (2π/Nθtest)
  ztest = f.(θtest)

  NN = 100;

  logNs = LinRange(log10(Nmin), log10(Nmax), NN)
  Ns = [floor(Int64, 10 ^ (logN)) for logN in logNs]
  errs = zeros(3, NN);

  for ii = 1:NN
    N = Ns[ii]
    Nθ = (3*(4N+1))÷2
    θ = (0:Nθ-1) .* (2π/Nθ)
    z = f.(θ);
    coef = lls_Fourier(θ, z, N)

    AF = FourierMat(θtest, 2N)
    errs[3, ii] = norm(AF*coef-ztest)/sqrt(Nθtest)

    rl = lls_RatFourier(θ, z, N)
    errs[1, ii] = norm(eval(rl, θtest)-ztest)/sqrt(Nθtest)

    rnl = nls_RatFourier(θ, z, N)
    errs[2, ii] = norm(eval(rnl, θtest)-ztest)/sqrt(Nθtest)

    println("ii=$(ii), N = $N"); flush(stdout)
  end

  return Ns, errs
end
